import ipaddress
import time
from pprint import pprint
from threading import Thread

from vrouter.client import ExtendedVPPAPIClient, BasicVPPAPIClient
from vrouter.common import models


def simple_example():
    with ExtendedVPPAPIClient() as client:
        # reply = client.add_configuration(
        #     # adding configuration in wrong order just to see if dependencies work
        #     models.RouteConfigurationItem(
        #         destination_network="10.1.3.0/24",
        #         next_hop_address="192.0.3.1",
        #         outgoing_interface="my-tap"
        #     )
        # )
        # pprint(reply)
        reply = client.add_configuration(
            models.InterfaceConfigurationItem(
                type=models.InterfaceType.TAP,
                name="GigabitEthernet3/0/0",
                enabled=True,
                ip_addresses=["10.100.0.1/24"],
            )
        )
        pprint(reply)

    # Python VPP API distributed with VPP (non-extended/original PAPI)
    with BasicVPPAPIClient() as client:
        pass
        # rv = client.tap_create_v2()
        # tap_sw_index = rv[3]  # rv.sw_if_index
        # pprint(rv)
        #
        # rv = client.sw_interface_set_flags(
        #     sw_if_index=tap_sw_index,
        #     flags=1  # VppEnum.vl_api_if_status_flags_t.IF_STATUS_API_FLAG_LINK_UP
        # )
        # pprint(rv)
        #
        # fibpaths = None
        # prefix_addr = IPv4Network('10.1.3.0/24')
        # next_hop_addr = IPv4Address('192.0.3.1')
        # rv = client.ip_route_add_del(
        #     is_add=1,
        #     is_multipath=0,
        #     route={'table_id': 0,
        #            'prefix': prefix_addr,
        #            # 'src': self.src,
        #            'n_paths': 1,
        #            'paths': [VppRoutePath(next_hop_addr, tap_sw_index).encode()],
        #            # 0xffffffff for no interface sw index
        #            },
        # )
        #
        # pprint(rv)
        #
        # for intf in client.sw_interface_dump():
        #     print(intf)
        #     # print(intf.interface_name)


def duplicated_config_example():
    with ExtendedVPPAPIClient() as client:
        interface_config = models.InterfaceConfigurationItem(
            name="my-tap", enabled=True, ip_addresses=["10.0.1.0/24"]
        )
        reply = client.add_configuration(interface_config)
        pprint(reply)
        assert reply.vpp_apply_success
        assert len(reply.vpp_apply_attempted_items) > 0
        assert not reply.processing_error

        reply = client.add_configuration(interface_config)
        pprint(reply)
        assert reply.processing_error
        assert reply.added_items[0].config == interface_config


def get_current_config_example():
    with ExtendedVPPAPIClient() as client:
        reply = client.add_configuration(
            # adding configuration in wrong order just to see if dependencies work
            models.RouteConfigurationItem(
                destination_network="10.1.3.0/24",
                next_hop_address="192.0.3.1",
                outgoing_interface="my-tap",
            )
        )
        pprint(reply)
        reply = client.get_configuration()
        pprint(reply)


def wait_with_context_example():
    # it is used for logical grouping of configuration here (grouping on client's user side),
    # but could hold any context information
    context = "1"

    def add_route(delay):
        time.sleep(delay)
        print("adding route")
        with ExtendedVPPAPIClient() as client:
            reply = client.add_configuration(
                # adding configuration in wrong order just to see if dependencies work
                models.RouteConfigurationItem(
                    destination_network="10.1.3.0/24",
                    next_hop_address="192.0.3.1",
                    outgoing_interface="my-tap",
                ),
                context=context,
            )
            pprint(reply)
            time.sleep(
                5
            )  # checking that running multiple clients from multiple threads doesn't raise Error
        pass

    def add_interface(delay):
        time.sleep(delay)
        print("adding interface")
        with ExtendedVPPAPIClient() as client:
            reply = client.add_configuration(
                models.InterfaceConfigurationItem(
                    name="my-tap", enabled=True, ip_addresses=["10.0.1.0/24"]
                ),
                context=context,
            )
            pprint(reply)
            time.sleep(
                5
            )  # checking that running multiple clients from multiple threads doesn't raise Error

    def wait_for_config(delay):
        time.sleep(delay)
        print("waiting for config")
        with ExtendedVPPAPIClient() as client:

            def check_application_to_vpp(config_item_details) -> bool:
                applied_config_items_with_right_context = [
                    detail
                    for detail in config_item_details
                    if detail.metadata.get("context") == context
                    and detail.state == models.State.APPLIED
                ]
                # NOTE: filter by context is not enough as waiting thread start sooner than all context configuration
                # is sent to vpp_client -> need to have more knowledge to wait properly (config item count in this case)
                return len(applied_config_items_with_right_context) == 2

            client.wait_for_configuration(configuration_check=check_application_to_vpp)
            print("config is applied in VPP")
        pass

    threads = [
        Thread(name="Thread-wait", target=wait_for_config, args=(1,)),
        Thread(name="Thread-route", target=add_route, args=(2,)),
        Thread(name="Thread-interface", target=add_interface, args=(3,)),
    ]
    try:
        for t in threads:
            t.start()
    except Exception as e:
        print("Error: unable to start thread due to ", e)
    try:
        for t in threads:
            t.join()
    except Exception as e:
        print("Error: unable to join thread due to ", e)


def route_delete_example():
    with ExtendedVPPAPIClient() as client:
        interface_config = models.InterfaceConfigurationItem(
            name="my-tap", enabled=True, ip_addresses=["10.0.1.0/24"]
        )
        route_config = models.RouteConfigurationItem(
            destination_network="10.1.3.0/24",
            next_hop_address="192.0.3.1",
            outgoing_interface="my-tap",
        )
        reply = client.add_configuration(interface_config, route_config)
        pprint(reply)
        reply = client.delete_configuration(route_config)
        pprint(reply)


def acl_example():
    with ExtendedVPPAPIClient() as client:
        # create ACL (with linked interfaces)
        reply = client.add_configuration(
            models.InterfaceConfigurationItem(
                name="acl-input-tap", enabled=True, ip_addresses=["10.100.1.1/24"]
            ),
            models.InterfaceConfigurationItem(
                name="acl-output-tap", enabled=True, ip_addresses=["10.200.1.1/24"]
            ),
            models.ACLConfigurationItem(
                name="ACL1",
                ingress=["acl-input-tap"],
                egress=["acl-output-tap"],
                rules=[
                    models.ACLRuleConfigurationItem(
                        action=models.ACLAction.PERMIT,
                        refinement=models.IPSpecification(
                            addresses=models.IPAddresses(
                                destination_network=ipaddress.IPv4Network(
                                    "10.200.1.0/24"
                                ),
                                source_network=ipaddress.IPv4Network("10.100.1.0/24"),
                            ),
                            protocol=models.ICMPProtocol(
                                icmp_code_range=models.ICMPRange(
                                    first=0, last=8
                                )  # includes echo and echo reply
                            ),
                        ),
                    ),
                    models.ACLRuleConfigurationItem(  # Deny all other
                        action=models.ACLAction.DENY
                    ),
                ],
            ),
        )
        pprint(reply)

        # delete existing ACL
        reply = client.get_configuration()
        acl_config = [
            item.config
            for item in reply.items
            if isinstance(item.config, models.ACLConfigurationItem)
            and item.config.name == "ACL1"
        ][0]
        reply = client.delete_configuration(acl_config)
        pprint(reply)


def nat_example():
    with ExtendedVPPAPIClient() as client:
        tap_out_config = models.InterfaceConfigurationItem(
            name="tap-nat-out", enabled=True, ip_addresses=["10.100.1.1/24"]
        )
        tap_in_config = models.InterfaceConfigurationItem(
            name="tap-nat-in", enabled=True, ip_addresses=["10.200.1.1/24"]
        )
        tap_out_nat_config = models.Nat44InterfaceConfigurationItem(
            name=tap_out_config.name, nat_outside=True
        )
        tap_in_nat_config = models.Nat44InterfaceConfigurationItem(
            name=tap_in_config.name, nat_inside=True
        )
        tcp_service_config = models.DNat44ConfigurationItem(
            label="tcp-service",
            static_mappings=[
                models.StaticMapping(
                    external_ip=ipaddress.IPv4Address("80.80.80.80"),
                    external_port=8888,
                    protocol=models.ProtocolInNAT.TCP,
                    local_ips=[
                        models.LocalIP(
                            local_ip=ipaddress.IPv4Interface(
                                tap_in_config.ip_addresses[0]
                            ).ip,
                            local_port=8000,
                        )
                    ],
                )
            ],
        )
        reply = client.add_configuration(
            tcp_service_config,
            tap_out_config,
            tap_in_config,
            tap_out_nat_config,
            tap_in_nat_config,
        )
        pprint(reply)

        # reply = client.delete_configuration(
        #     tcp_service_config, tap_in_nat_config, tap_out_nat_config
        # )
        # pprint(reply)


def wireguard_example():
    with ExtendedVPPAPIClient() as client:
        wireguard_interface_config = models.InterfaceConfigurationItem(
            name="my-wg",
            enabled=True,
            ip_addresses=["172.32.0.10/30"],
            type=models.InterfaceType.WIREGUARD_TUNNEL,
            link=models.WireguardInterfaceLink(
                private_key="WOysjzDd3oGVhi/FAXZTMUwFbWIb0xzjEQtQh30LhVg=",  # encoded in base64
                port=50869,
                src_addr=ipaddress.IPv4Address("10.10.1.1"),
            ),
        )
        wireguard_peer_config = models.WireguardPeerConfigurationItem(
            public_key="CV+HC1xgxWSojYJ0h8Jh1JxF7Lr3QL8WFmJjCrJzfGc=",  # encoded in base64
            port=50869,
            endpoint=ipaddress.IPv4Address("10.10.1.2"),
            allowed_ips=["0.0.0.0/0"],
            wg_if_name=wireguard_interface_config.name,
        )

        wireguard_route_config = models.RouteConfigurationItem(
            destination_network="172.32.0.9/32", outgoing_interface="my-wg"
        )
        reply = client.add_configuration(
            wireguard_interface_config, wireguard_peer_config, wireguard_route_config
        )
        print(reply)

        # reply = client.delete_configuration(wireguard_interface_config, wireguard_peer_config)
        # pprint(reply)


def frr_example():
    # Note: connection to FRR must be configured so this examples work (launch PAPI++ server with correctly filled
    # --frr_agent_host and --frr_agent_port parameters referring to running FRR agent (in docker container) )
    with ExtendedVPPAPIClient() as client:
        # PAPI++ server gets the default initial FRR configuration at startup -> ideal start for FRR config addition
        frr_init_config = [
            config_detail.config
            for config_detail in client.get_configuration().items
            if isinstance(config_detail.config, models.FRRConfigurationItem)
        ][0]
        pprint(frr_init_config)

        # appending some bgp configuration to initial FRR config
        # (FRR config is updated, but that means just means delete of old config and add of new one. In case of FRR,
        # the delete does nothing because FRR need to have always some configuration. That means that only addition of
        # FRR configuration matters.)
        append_to_frr_config = (
            """ 
router bgp 100
 !
 address-family ipv4 unicast
  distance bgp 100 100 100
 exit-address-family
exit
!       
        """.strip(
                "\n "
            )
            + "\n"
        )
        reply = client.update_configuration(
            [frr_init_config],
            [
                models.FRRConfigurationItem(
                    config=frr_init_config.config + append_to_frr_config
                )
            ],
        )
        pprint(reply, width=260)


def failover_notification():
    # initial topology
    # (2 data interfaces representing 2 redundant links into core network -> failover means switching between them)
    name_of_data_interface1 = "data-interface1"
    name_of_data_interface2 = "data-interface2"
    with ExtendedVPPAPIClient() as client:
        client.add_configuration(
            models.InterfaceConfigurationItem(
                name=name_of_data_interface1, enabled=True, ip_addresses=["10.0.1.1/24"]
            ),
            models.InterfaceConfigurationItem(
                name=name_of_data_interface2, enabled=True, ip_addresses=["10.0.2.1/24"]
            ),
            models.RouteConfigurationItem(
                destination_network="10.100.200.0/24",
                next_hop_address="10.0.1.2",
                outgoing_interface=name_of_data_interface1,
            ),
        )

    # define 2 threads, one simulating switching between data links in case of failover and other the client being
    # notified about it:
    def failover_switcher(delay):
        """Simulates FRR's OSPF link down detection.

        When data link is down, the OSPF will learn new routes that will reroute traffic to the other data link that is
        still functional and remove routes routing packets to data link that is down."""
        time.sleep(delay)
        print("failover switcher started")

        def failover_to_data_link2(client):
            # remove routes to link1
            client.delete_configuration(
                models.RouteConfigurationItem(
                    destination_network="10.100.200.0/24",
                    next_hop_address="10.0.1.2",
                    outgoing_interface=name_of_data_interface1,
                )
            )
            # replace old routes by adding the same routes, but to link2
            client.add_configuration(
                models.RouteConfigurationItem(
                    destination_network="10.100.200.0/24",
                    next_hop_address="10.0.2.2",
                    outgoing_interface=name_of_data_interface2,
                )
            )

        def failover_to_data_link1(client):
            # remove routes to link2
            client.delete_configuration(
                models.RouteConfigurationItem(
                    destination_network="10.100.200.0/24",
                    next_hop_address="10.0.2.2",
                    outgoing_interface=name_of_data_interface2,
                )
            )
            # replace old routes by adding the same routes, but to link1
            client.add_configuration(
                models.RouteConfigurationItem(
                    destination_network="10.100.200.0/24",
                    next_hop_address="10.0.1.2",
                    outgoing_interface=name_of_data_interface1,
                )
            )

        with ExtendedVPPAPIClient() as client:
            time.sleep(5)
            failover_to_data_link2(client)
            time.sleep(5)
            failover_to_data_link1(client)

    def notified_client(delay):
        """Client that will be "notified" about the failover.

        The notification means to actively check for certain route changes in PAPI++ server configuration
        """
        time.sleep(delay)
        print("notified client started")

        def wait_for_failover(client, link1, link2):
            print(client.get_configuration().items)
            initial_data_link = [
                detail.config.outgoing_interface
                for detail in client.get_configuration().items
                if detail.state == models.State.APPLIED
                and isinstance(detail.config, models.RouteConfigurationItem)
                and detail.config.outgoing_interface in [link1, link2]
            ][0]

            def check_for_failover(config_item_details) -> bool:
                actual_data_link = [
                    detail.config.outgoing_interface
                    for detail in config_item_details
                    if detail.state == models.State.APPLIED
                    and isinstance(detail.config, models.RouteConfigurationItem)
                    and detail.config.outgoing_interface in [link1, link2]
                ][0]
                return initial_data_link != actual_data_link

            client.wait_for_configuration(configuration_check=check_for_failover)
            if initial_data_link == link1:
                return link1, link2
            else:
                return link2, link1

        with ExtendedVPPAPIClient() as client:
            failing_link, new_link = wait_for_failover(
                client, name_of_data_interface1, name_of_data_interface2
            )
            print(f"detected failover from {failing_link} to {new_link}")
            failing_link, new_link = wait_for_failover(
                client, name_of_data_interface1, name_of_data_interface2
            )
            print(f"detected failover from {failing_link} to {new_link}")

    threads = [
        Thread(name="Thread-notified-client", target=notified_client, args=(1,)),
        Thread(name="Thread-failover-switcher", target=failover_switcher, args=(2,)),
    ]
    try:
        for t in threads:
            t.start()
    except Exception as e:
        print("Error: unable to start thread due to ", e)
    try:
        for t in threads:
            t.join()
    except Exception as e:
        print("Error: unable to join thread due to ", e)


def update_dpdk_interface():
    """Example showing how to update state and ip address of DPDK interface(interface not created by API, but created
    automatically from VPP's startup configuration)."""

    # Note: for example to work, you need to configure VPP with one dpdk interface in startup.conf and name it
    # "my-dpdk-interface" (see ../dev-setup/startup.conf)
    with ExtendedVPPAPIClient() as client:
        dpdk_init_config = [
            item.config
            for item in client.get_configuration().items
            if item.config.name == "GigabitEthernet3/0/0"
        ][0]
        reply = client.update_configuration(
            dpdk_init_config,
            models.InterfaceConfigurationItem(
                name=dpdk_init_config.name,
                type=dpdk_init_config.type,
                enabled=True,
                ip_addresses=["10.10.1.1/24"],
            ),
        )
        pprint(reply)


if __name__ == "__main__":
    # simple_example()
    # duplicated_config_example()
    # get_current_config_example()
    # wait_with_context_example()
    # route_delete_example()
    # acl_example()
    # nat_example()
    # wireguard_example()
    # frr_example()
    # failover_notification()

    update_dpdk_interface()

    # wireguard_example()
    # nat_example()



"""
Client for VPP configuration server
"""

import time
from typing import Callable, List

from vrouter.common.utils import configure_pyro5_logging

configure_pyro5_logging()  # must be before any Pyro5 import to configure logging
from Pyro5.api import Proxy

from vrouter.common import serializers
from vrouter.common.models import ConfigurationItemDetail


class ExtendedVPPAPIClient(Proxy):
    """Python VPP client based on fd.io VPP client, but with new customized API and additional capabilities.

    This is the basic entry point for configuration using this repository as python VPP configuration solution
    """

    def __init__(self, server_host: str = "0.0.0.0", server_port: int = 9999):
        """Creates and initializes extended VPP client

        :param server_host: hostname/ip address of the Pyro5 nameserver
        :param server_port: port of the Pyro5 nameserver
        """
        serializers.import_pyro5_serializers()  # fix for serializing of custom types
        super().__init__(f"PYRONAME:extended.papi@{server_host}:{server_port}")

    def wait_for_configuration(
        self,
        configuration_check: Callable[[List[ConfigurationItemDetail]], bool],
        retry_count: int = -1,
        delay: float = 0.5,
    ):
        """Waits for configuration to pass the provided configuration check"""
        while retry_count > 0 or retry_count <= -1:
            if configuration_check(self.get_configuration().items):
                return
            time.sleep(delay)
            retry_count = retry_count - 1


class BasicVPPAPIClient(Proxy):
    """Wrapper for original fd.io VPP python client that should have provide the same functionality with the difference
    of using client/server architecture (the Pyro5 library).

    Currently the problem is the serialization of the data sending to/receiving from server that should configure VPP.
    The input/output data from fd.io paython client is not serializable and is received/sent through Pyro5 library that
    uses serialization in garbled state (problem is mostly with the fd.io python client response)
    """

    def __init__(self):
        serializers.import_pyro5_serializers()  # fix for serializing of custom types
        super().__init__("PYRONAME:papi")



"""
Serialization utilities/fixes for Pyro5 serialization of data structures flowing between client and server
"""

import ipaddress
import typing
import warnings
from dataclasses import is_dataclass, fields

import Pyro5.api
from vpp_papi import macaddress, vpp_serializer

from vrouter.common import models

PYRO_SERIALIZED_CLASS_NAME_KEY = "__class__"
SERIALIZED_CLASS_VALUE_KEY = "json"


def convert_mac_dict_to_class(class_name, mac_dict):
    return macaddress.MACAddress(mac_dict["mac"])


def convert_mac_class_to_dict(mac_class):
    return {
        PYRO_SERIALIZED_CLASS_NAME_KEY: "vpp_papi.macaddress.MACAddress",
        "mac": str(mac_class),
    }


def convert_ipv4address_dict_to_class(class_name, ipv4addr_dict):
    return ipaddress.IPv4Address(ipv4addr_dict["ipv4_address"])


def convert_ipv4address_class_to_dict(ipv4addr_class):
    return {
        PYRO_SERIALIZED_CLASS_NAME_KEY: "ipaddress.IPv4Address",
        "ipv4_address": ipv4addr_class.exploded,
    }


def convert_ipv4network_dict_to_class(class_name, ipv4net_dict):
    return ipaddress.IPv4Network(ipv4net_dict["ipv4_network"])


def convert_ipv4network_class_to_dict(ipv4net_class):
    return {
        PYRO_SERIALIZED_CLASS_NAME_KEY: "ipaddress.IPv4Network",
        "ipv4_network": ipv4net_class.exploded,
    }


def convert_vpp_serializer_value_error_dict_to_class(class_name, error_dict):
    return vpp_serializer.VPPSerializerValueError(error_dict["VPPSerializerValueError"])


def convert_vpp_serializer_value_error_class_to_dict(clazz):
    return {
        PYRO_SERIALIZED_CLASS_NAME_KEY: "vpp_papi.vpp_serializer.VPPSerializerValueError",
        "VPPSerializerValueError": str(clazz),
    }


dataclass_name_to_class = dict()
nondataclass_name_to_dict_to_class_function = dict()


def register_serialization_for_dataclass(class_name, dataclass):
    """Custom serialization for dataclasses"""

    dataclass_name_to_class[class_name] = dataclass
    Pyro5.api.SerializerBase.register_class_to_dict(
        dataclass,
        lambda dataclass_instance: {
            PYRO_SERIALIZED_CLASS_NAME_KEY: class_name,
            SERIALIZED_CLASS_VALUE_KEY: dataclass.schema().dump(dataclass_instance),
        },
    )
    Pyro5.api.SerializerBase.register_dict_to_class(
        class_name,
        lambda class_name, dataclass_dict: dataclass_inner_deserialization(
            dataclass.schema().load(dataclass_dict[SERIALIZED_CLASS_VALUE_KEY])
        ),
    )


def register_serialization_for_nondataclass(clazz, name, class_to_dict, dict_to_class):
    """Custom serialization for other classes than dataclasses"""

    nondataclass_name_to_dict_to_class_function[name] = dict_to_class
    Pyro5.api.SerializerBase.register_class_to_dict(clazz, class_to_dict)
    Pyro5.api.SerializerBase.register_dict_to_class(name, dict_to_class)
    # ignore dataclass-json converter warning because dataclass_inner_deserialization(...) will be used
    # to resolve types
    warnings.filterwarnings(
        action="ignore",
        message=".*{}.*It's advised to pass the correct marshmallow type.*".format(
            name.replace(".", "\\.")
        ),
    )


def import_pyro5_serializers():
    """Fixes Pyro5 serialization for client-server traffic by registering customized serializers"""

    # configuration items
    register_serialization_for_dataclass(
        "models.InterfaceConfigurationItem", models.InterfaceConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.RouteConfigurationItem", models.RouteConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.ACLConfigurationItem", models.ACLConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.Nat44AddressPoolConfigurationItem",
        models.Nat44AddressPoolConfigurationItem,
    )
    register_serialization_for_dataclass(
        "models.Nat44InterfaceConfigurationItem", models.Nat44InterfaceConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.DNat44ConfigurationItem", models.DNat44ConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.WireguardPeerConfigurationItem", models.WireguardPeerConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.LCPGlobalsConfigurationItem", models.LCPGlobalsConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.LCPPairConfigurationItem", models.LCPPairConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.FRRConfigurationItem", models.FRRConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.DHCPConfigurationItem", models.DHCPConfigurationItem
    )
    register_serialization_for_dataclass(
        "models.RunningPacketCapture", models.RunningPacketCapture
    )
    register_serialization_for_dataclass(
        "models.LinuxDockerVethTunnelConfigurationItem",
        models.LinuxDockerVethTunnelConfigurationItem,
    )
    register_serialization_for_dataclass(
        "models.LinuxRouteConfigurationItem", models.LinuxRouteConfigurationItem
    )

    # replies
    register_serialization_for_dataclass("models.ApplyReply", models.ApplyReply)
    register_serialization_for_dataclass(
        "models.AddConfigurationItemsReply", models.AddConfigurationItemsReply
    )
    register_serialization_for_dataclass(
        "models.DeleteConfigurationItemsReply", models.DeleteConfigurationItemsReply
    )
    register_serialization_for_dataclass(
        "models.UpdateConfigurationItemsReply", models.UpdateConfigurationItemsReply
    )
    register_serialization_for_dataclass(
        "models.GetConfigurationItemsReply", models.GetConfigurationItemsReply
    )
    register_serialization_for_dataclass(
        "models.ConfigurationItemReply", models.ConfigurationItemReply
    )
    register_serialization_for_dataclass(
        "models.ConfigurationItemDetail", models.ConfigurationItemDetail
    )
    register_serialization_for_dataclass(
        "models.WireguardPeerDetail", models.WireguardPeerDetail
    )
    register_serialization_for_dataclass("models.LCPPairDetail", models.LCPPairDetail)
    register_serialization_for_dataclass(
        "models.ProcessCallReply", models.ProcessCallReply
    )

    # static models for dynamic api object from basic vpp client
    # (pyro5 handles serialization, modifier function using dataclass-json does deserialization)
    register_serialization_for_dataclass(
        "models.SWInterfaceDetail", models.SWInterfaceDetail
    )
    register_serialization_for_dataclass("models.IPRouteDetail", models.IPRouteDetail)
    register_serialization_for_dataclass("models.ACLRule", models.ACLRule)
    register_serialization_for_dataclass("models.ACLDetails", models.ACLDetails)
    register_serialization_for_dataclass(
        "models.NAT44AddressDetails", models.NAT44AddressDetails
    )
    register_serialization_for_dataclass(
        "models.NAT44InterfaceDetails", models.NAT44InterfaceDetails
    )
    register_serialization_for_dataclass(
        "models.NAT44StaticMappingDetails", models.NAT44StaticMappingDetails
    )
    register_serialization_for_dataclass(
        "models.NAT44IdentityMappingDetails", models.NAT44IdentityMappingDetails
    )

    # other objects that need serialization
    register_serialization_for_nondataclass(
        macaddress.MACAddress,
        "vpp_papi.macaddress.MACAddress",
        convert_mac_class_to_dict,
        convert_mac_dict_to_class,
    )
    register_serialization_for_nondataclass(
        vpp_serializer.VPPSerializerValueError,
        "vpp_papi.vpp_serializer.VPPSerializerValueError",
        convert_vpp_serializer_value_error_class_to_dict,
        convert_vpp_serializer_value_error_dict_to_class,
    )
    register_serialization_for_nondataclass(
        ipaddress.IPv4Address,
        "ipaddress.IPv4Address",
        convert_ipv4address_class_to_dict,
        convert_ipv4address_dict_to_class,
    )
    register_serialization_for_nondataclass(
        ipaddress.IPv4Network,
        "ipaddress.IPv4Network",
        convert_ipv4network_class_to_dict,
        convert_ipv4network_dict_to_class,
    )


def dataclass_inner_deserialization(data):
    if type(data) is list:
        return [dataclass_inner_deserialization(e) for e in data]

    if type(data) is dict:
        return _handle_dict_value(data)

    if not is_dataclass(data):
        return data

    for field in fields(type(data)):
        is_opt = _is_optional_or_union_with_none(field.type)
        if is_opt:
            field_value = getattr(data, field.name, None)
            if field_value is None:
                continue
        else:
            field_value = getattr(data, field.name)

        if is_dataclass(field_value):
            setattr(data, field.name, dataclass_inner_deserialization(field_value))

        if type(field_value) is list:
            setattr(
                data,
                field.name,
                [dataclass_inner_deserialization(e) for e in field_value],
            )
            continue

        if type(field_value) is dict:
            setattr(data, field.name, _handle_dict_value(field_value))
            continue

        # do nothing for other types that we don't deserialize (literals, other types...)

    return data


def _is_optional_or_union_with_none(T):
    origin_type = getattr(T, "__origin__", None)
    type_args = getattr(T, "__args__", [])
    return origin_type is typing.Union and type_args and type_args[-1] is type(None)


def _handle_dict_value(dict_value):
    to_deserialize_class_name = dict_value.get(PYRO_SERIALIZED_CLASS_NAME_KEY)
    if to_deserialize_class_name is not None:
        clazz = dataclass_name_to_class.get(to_deserialize_class_name)
        if clazz is not None:
            deserialized_value = clazz.schema().load(
                dict_value.get(SERIALIZED_CLASS_VALUE_KEY)
            )  # still can have inner not deserialized dataclasses
        else:
            dict_to_class_function = nondataclass_name_to_dict_to_class_function.get(
                to_deserialize_class_name
            )
            deserialized_value = dict_to_class_function(
                to_deserialize_class_name, dict_value
            )
        value = dataclass_inner_deserialization(deserialized_value)
    else:
        value = {k: dataclass_inner_deserialization(e) for k, e in dict_value.items()}
    return value



"""
The agent running inside FRR container. The server relays FRR-related commands to this agent and the agent performs then
on the FRR instance. The agent also copy BGP-learned routes into VPP by acting as configuration client configuring routes
using server. The connection between agent and server is done by using Pyro5, just like in client/server connection.
"""

import logging
import subprocess
import threading

from vrouter.common.models import ProcessCallReply
from vrouter.common.utils import configure_main_logger, configure_pyro5_logging

logger = configure_main_logger(f"frr-agent", logging.DEBUG)
configure_pyro5_logging()  # must be before any Pyro5 import to configure logging
from vrouter.server import pyro
from vrouter.frr import route_reflector


def frr_api_wrapper():
    """Creates wrapper class that has all FRR API methods. The wrapper represents FRR API server"""

    @pyro.api_expose
    class Wrapper(pyro.NamedWrapper):
        def __init__(self):
            self.lock = threading.Lock()

        @staticmethod
        def name():
            return "frr.api"

        def run_vtysh_command(self, command) -> ProcessCallReply:
            try:
                self.lock.acquire(blocking=True)
                result = subprocess.run(
                    f"""vtysh -c "{command}" """, shell=True, capture_output=True
                )
                logger.info(f"Call of run_vtysh_command: {command} (result: {result})")
                return ProcessCallReply(
                    return_code=result.returncode,
                    stdout=result.stdout,
                    stderr=result.stderr,
                )
            except Exception:
                logger.exception("Exception in run_vtysh_command")
            finally:
                self.lock.release()

        def load_vtysh_configuration(self, configuration: str) -> ProcessCallReply:
            try:
                self.lock.acquire(blocking=True)
                with open(r"/etc/frr/frr.conf", "w") as frr_config:
                    frr_config.write(configuration)
                result = subprocess.run("vtysh -b", shell=True, capture_output=True)
                logger.info(
                    f"Call of load_vtysh_configuration: configuration:\n{configuration}\n (result: {result})"
                )
                return ProcessCallReply(
                    return_code=result.returncode,
                    stdout=result.stdout,
                    stderr=result.stderr,
                )
            except Exception:
                logger.exception("Exception in load_vtysh_configuration")
            finally:
                self.lock.release()

        def configure_route_reflector(
            self, papi_server_host: str = "0.0.0.0", papi_server_port: int = 9999
        ):
            try:
                self.lock.acquire(blocking=True)
                logger.info(
                    f"configuring route reflector (papi_server_host={papi_server_host}, "
                    f"papi_server_port={papi_server_port})"
                )
                route_reflector.configure_route_reflector(
                    papi_server_host=papi_server_host, papi_server_port=papi_server_port
                )
            except Exception:
                logger.exception("Exception in configure_route_reflector")
            finally:
                self.lock.release()

    return Wrapper


if __name__ == "__main__":
    graceful_shutdown_callback = route_reflector.start_route_reflector_in_background()
    pyro.run_server_loop(
        frr_api_wrapper(), pyro_daemon_port=6666, nameserver_daemon_port=7777
    )
    graceful_shutdown_callback()



"""
Server that configures VPP
"""

import fnmatch
import logging
import os
import threading

from vrouter.common.utils import configure_main_logger, configure_pyro5_logging

configure_pyro5_logging()  # must be before any Pyro5 import to configure logging
logger = configure_main_logger("papi-server", logging.INFO)

from vpp_papi import VPPApiClient, VPPIOError

from vrouter.server import __version__, pyro
from vrouter.common.models import (
    SW_IF_INDEX,
    NO_INTERFACE_SW_INDEX,
    InterfaceType,
    InterfaceConfigurationItem,
    ApplyReply,
    State,
    AddConfigurationItemsReply,
    ConfigurationItemReply,
    ConfigurationItemDetail,
    GetConfigurationItemsReply,
    VPP_DUMP_KEY,
    DeleteConfigurationItemsReply,
    ACLConfigurationItem,
    ACL_INDEX,
    NO_ACL_INDEX,
    WireguardPeerConfigurationItem,
    WIREGUARD_PEER_INDEX,
    NO_WIREGUARD_PEER_INDEX,
    VPPConfigUpdater,
    UpdateConfigurationItemsReply,
    RunningPacketCapture,
    IF_STATUS_API_FLAG_ADMIN_UP,
)
from vrouter.server.handlers import (
    InterfaceHandler,
    RouteHandler,
    ACLHandler,
    Nat44AddressPoolHandler,
    Nat44InterfaceHandler,
    DNat44Handler,
    WireguardPeerHandler,
    FRRHandler,
    FRRAgentClient,
    LCPHandler,
    RunningPacketCaptureHandler,
    LinuxDockerVethTunnelHandler,
    LinuxRouteHandler,
    DHCPHandler,
)


def _get_vpp_api_json_files(vpp_json_root_dir):
    """Get filenames of VPP API JSON files"""
    jsonfiles = []

    for root, dirnames, filenames in os.walk(vpp_json_root_dir):
        for filename in fnmatch.filter(filenames, "*.api.json"):
            jsonfiles.append(os.path.join(root, filename))

    if not jsonfiles:
        logger.error("no json api files found")
        exit(-1)
    return jsonfiles


class ExtendedVPPClient:
    """Client based on fd.io VPP client that have extended capabilties.

    Don't interchange it with client.ExtendedVPPAPIClient. This client is extension of fd.io VPP client and serves as
    main implementation of Pyro5 server. This is where client.ExtendedVPPAPIClient to call end
    """

    def __init__(
        self,
        vpp_api_json_root_dir,
        vpp_socket_full_path,
        frr_agent_host="0.0.0.0",
        frr_agent_port=7777,
        dhcp_socket="/run/kea/control_socket_4",
    ):
        self._init_inner_vpp_client(vpp_api_json_root_dir, vpp_socket_full_path)
        self._frr_agent_host = frr_agent_host
        self._frr_agent_port = frr_agent_port
        self._dhcp_socket = dhcp_socket
        self._register_handlers()
        self._configuration_item_records = set()
        self._config_init_from_vpp_dump()
        self._config_init_from_frr_dump()
        self._configure_route_reflector()
        self._transaction_counter = 0

    def _init_inner_vpp_client(self, vpp_api_json_root_dir, vpp_socket_full_path):
        """Setup fd.io client"""
        self.inner_vpp_client = VPPApiClient(
            apifiles=_get_vpp_api_json_files(vpp_api_json_root_dir),
            server_address=vpp_socket_full_path,
        )
        try:
            rv = self.inner_vpp_client.connect("extended_vpp_client")
        except VPPIOError as e:
            logger.exception(f"can't connect to VPP due to ${e}")
            exit(-1)
        if rv != 0:
            logger.error(f"can't connect to VPP (return value ${rv}")
            exit(-1)
        # TODO: where to put vpp_client.disconnect()?

    def _register_handlers(self):
        self.handlers = [
            InterfaceHandler(self.inner_vpp_client, self.get_sw_if_index),
            RouteHandler(self.inner_vpp_client, self.get_sw_if_index),
            ACLHandler(self.inner_vpp_client, self.get_sw_if_index, self.get_acl_index),
            Nat44AddressPoolHandler(self.inner_vpp_client),
            Nat44InterfaceHandler(self.inner_vpp_client, self.get_sw_if_index),
            DNat44Handler(self.inner_vpp_client, self.get_sw_if_index),
            WireguardPeerHandler(
                self.inner_vpp_client,
                self.get_sw_if_index,
                self.get_wireguard_peer_index,
            ),
            LCPHandler(self.inner_vpp_client, self.get_sw_if_index),
            RunningPacketCaptureHandler(
                self.inner_vpp_client,
                self.get_sw_if_index,
                self.get_current_running_packet_capture,
            ),
            LinuxDockerVethTunnelHandler(),
            LinuxRouteHandler(),
            DHCPHandler(self._dhcp_socket),
        ]
        if (
            self._frr_agent_host
        ):  # enable FRR only if FRR Agent address is configured (non-empty string)
            self.handlers.append(FRRHandler(self._frr_agent_host, self._frr_agent_port))

    def _config_init_from_vpp_dump(self):
        """Read state from VPP instance and convert it to server state (ConfigurationItemRecord-s)"""
        for interface_detail in self.inner_vpp_client.api.sw_interface_dump():
            configuration_item = InterfaceConfigurationItem(
                name=interface_detail.interface_name,
                type=InterfaceType.get_from_vpp_str(
                    interface_detail.interface_dev_type
                ),
                enabled=(interface_detail.flags & IF_STATUS_API_FLAG_ADMIN_UP) != 0,
                # ignoring ip_addresses
            )
            self._configuration_item_records.add(
                ConfigurationItemRecord(
                    configuration_item,
                    self._get_handler(configuration_item),
                    init_state=State.APPLIED,
                    metadata={SW_IF_INDEX: interface_detail.sw_if_index},
                )
            )

    def _config_init_from_frr_dump(self):
        """Read state from FRR instance and convert it to server state (ConfigurationItemRecord-s)"""
        if not self._frr_agent_host:  # disabling FRR if not configured properly
            return
        frr_handler = [
            handler for handler in self.handlers if isinstance(handler, FRRHandler)
        ][0]
        self._configuration_item_records.add(
            ConfigurationItemRecord(
                frr_handler.dump_from_vpp({}),
                frr_handler,
                init_state=State.APPLIED,
                metadata={},
            )
        )

    def _configure_route_reflector(self):
        """Inform route reflector in FRR's agent how to connect back to this server.

        FRR starts first and this server is configured with info how to connect to FRR agent. However, the FRR agent
        needs to connect to this server too (learned BGP route must be copied into VPP using this server). So this
        method informs FRR's agent how to do it."""

        if not self._frr_agent_host:  # disabling FRR if not configured properly
            return
        try:
            with FRRAgentClient(
                frr_agent_host=self._frr_agent_host, frr_agent_port=self._frr_agent_port
            ) as client:
                client.configure_route_reflector(
                    papi_server_host=pyro.get_external_ip_address(),
                    papi_server_port=9999,
                )
        except Exception:
            logger.exception("can't configure FRR's route reflector")

    def get_sw_if_index(
        self, interface_name
    ):  # FIXME get this info somehow without compromising vpp_client API
        for record in self._configuration_item_records:
            if (
                isinstance(record.configuration_item, InterfaceConfigurationItem)
                and record.state == State.APPLIED
                and record.configuration_item.name == interface_name
            ):
                return record.metadata[SW_IF_INDEX]
        return NO_INTERFACE_SW_INDEX  # FIXME Or raise error?

    def get_acl_index(
        self, acl_name
    ):  # FIXME get this info somehow without compromising vpp_client API
        for record in self._configuration_item_records:
            if (
                isinstance(record.configuration_item, ACLConfigurationItem)
                and record.state == State.APPLIED
                and record.configuration_item.name == acl_name
            ):
                return record.metadata[ACL_INDEX]
        return NO_ACL_INDEX  # FIXME Or raise error?

    # FIXME get this info somehow without compromising vpp_client API
    def get_wireguard_peer_index(self, wg_peer_config_item):
        for record in self._configuration_item_records:
            if (
                isinstance(record.configuration_item, WireguardPeerConfigurationItem)
                and record.state == State.APPLIED
                and record.configuration_item == wg_peer_config_item
            ):
                return record.metadata[WIREGUARD_PEER_INDEX]
        return NO_WIREGUARD_PEER_INDEX  # FIXME Or raise error?

    def get_current_running_packet_capture(self):
        for record in self._configuration_item_records:
            if (
                isinstance(record.configuration_item, RunningPacketCapture)
                and record.state == State.APPLIED
            ):
                return record.configuration_item
        return None

    def _get_handler(self, configuration_item):
        for handler in self.handlers:
            if handler.is_handling_configuration_item(configuration_item):
                return handler

    def add_configuration(
        self, *configuration_items, context: str = ""
    ) -> AddConfigurationItemsReply:
        """Server implementation for adding configuration items"""

        if not configuration_items:
            # nothing to apply, but that nothing is applied successfully
            return AddConfigurationItemsReply()

        # check duplicate config items
        duplicate_config_items = [
            record.get_configuration_item_reply()
            for configuration_item in configuration_items
            for record in self._configuration_item_records
            if configuration_item == record.configuration_item
        ]
        if (
            duplicate_config_items
        ):  # duplicate config items are expected to fail config item addition
            return AddConfigurationItemsReply(
                processing_error="Found duplicate config items.",
                all_added_items_applied_to_vpp=False,
                # TODO move duplicated items to processing error, make it Error instead of string?
                added_items=duplicate_config_items,
            )

        # update configuration
        added_config_record = []
        for configuration_item in configuration_items:
            metadata = {"context": context} if context else dict()
            new_record = ConfigurationItemRecord(
                configuration_item,
                self._get_handler(configuration_item),
                metadata=metadata,
            )
            added_config_record.append(new_record)
            self._configuration_item_records.add(new_record)

        # get all configuration that can be applied
        def get_what_can_be_directly_added_to_vpp():
            labels_of_applied_config = {
                label
                for conf_item in self._configuration_item_records
                if conf_item.state == State.APPLIED
                for label in conf_item.handler.get_labels(conf_item.configuration_item)
            }
            not_applied_conf_items = {
                conf_item
                for conf_item in self._configuration_item_records
                if conf_item.state
                not in [
                    State.APPLIED,
                    State.BLOCKED_BY_ADD_APPLY_FAILURE,
                    State.BLOCKED_BY_REMOVE_APPLY_FAILURE,
                ]
            }
            return [
                not_applied_conf_item
                for not_applied_conf_item in not_applied_conf_items
                if not_applied_conf_item.handler.get_dependency_labels(
                    not_applied_conf_item.configuration_item
                ).issubset(labels_of_applied_config)
            ]

        # applied configuration that is new/unblocked
        configuration_item_replies = []
        self._transaction_counter += 1
        ignore_for_blocked_records_output = set()
        # TODO rename transaction to "Config Application" ?
        #  (it is not really a transaction due to lack of rollback and final commit)
        logger.info(
            "{}\nTransaction {} started\n{}".format(
                "=" * 30, self._transaction_counter, "=" * 30
            )
        )
        while True:
            can_be_added = get_what_can_be_directly_added_to_vpp()
            if can_be_added:
                for index, to_add in enumerate(can_be_added):
                    try:
                        reply = to_add.handler.add_to_vpp(to_add.configuration_item)
                    except Exception as e:
                        logger.exception(
                            f"Exception raised while applying {to_add.configuration_item}"
                        )
                        reply = ApplyReply(success=False, error=repr(e))
                    if reply.success:
                        logger.info(
                            "\tSUCCESS: added to VPP config item {}".format(
                                repr(to_add.configuration_item)
                            )
                        )
                        to_add.state = State.APPLIED
                        to_add.metadata.update(reply.vpp_data)
                        configuration_item_replies.append(
                            to_add.get_configuration_item_reply()
                        )
                        try:
                            to_add.metadata[VPP_DUMP_KEY] = (
                                to_add.handler.dump_from_vpp(to_add.configuration_item)
                            )
                        except Exception as e:
                            logger.warning(
                                f"No dump data from VPP will be provided due to failure of dumping: {str(e)} "
                                + f"(dumping for configuration item {to_add.configuration_item})"
                            )
                    else:
                        logger.error(
                            "\tFAILED: adding to VPP config item {}\n\t\tREASON:{}".format(
                                repr(to_add.configuration_item), reply.error
                            )
                        )
                        to_add.state = State.BLOCKED_BY_ADD_APPLY_FAILURE
                        ignore_for_blocked_records_output.add(to_add)
                        configuration_item_replies.append(
                            to_add.get_configuration_item_reply(error=reply.error)
                        )
                        # TODO reapply of failing config? with some retry limit?
            else:
                # change state for new config items not picked for applying (=> they are blocked by something)
                for conf_item in self._configuration_item_records:
                    if conf_item.state == State.RECEIVED:
                        conf_item.state = State.BLOCKED

                self._print_blocked_records(
                    ignore_records=ignore_for_blocked_records_output
                )
                break
        logger.info(
            "{}\nTransaction {} ended\n{}\n".format(
                "=" * 30, self._transaction_counter, "=" * 30
            )
        )

        return AddConfigurationItemsReply(
            all_added_items_applied_to_vpp=not (
                False
                in [record.state == State.APPLIED for record in added_config_record]
            ),
            added_items=[
                record.get_configuration_item_detail() for record in added_config_record
            ],
            vpp_apply_success=not (
                True in [bool(reply.error) for reply in configuration_item_replies]
            ),
            vpp_apply_attempted_items=configuration_item_replies,
        )

    def _print_blocked_records(self, ignore_records=set()):
        """Retrieve and print configuration items that are for some reason blocked and prints them (as part
        of the transaction)"""

        labels_of_applied_config = {
            label
            for conf_item in self._configuration_item_records
            if conf_item.state == State.APPLIED
            for label in conf_item.handler.get_labels(conf_item.configuration_item)
        }
        for record in self._configuration_item_records.difference(ignore_records):
            match record.state:
                case State.BLOCKED:
                    logger.warning(
                        "\tBLOCKED: {}\n\t\tBY:{}".format(
                            repr(record.configuration_item),
                            repr(
                                record.handler.get_dependency_labels(
                                    record.configuration_item
                                ).difference(labels_of_applied_config)
                            ),
                        )
                    )
                case State.BLOCKED_BY_ADD_APPLY_FAILURE:
                    logger.warning(
                        "\tBLOCKED: {}\n\t\tBY: Transaction failure (while adding config)".format(
                            repr(record.configuration_item)
                        )
                    )
                case State.BLOCKED_BY_REMOVE_APPLY_FAILURE:
                    logger.warning(
                        "\tBLOCKED: {}\n\t\tBY: Transaction failure (while removing config)".format(
                            repr(record.configuration_item)
                        )
                    )

    def delete_configuration(
        self, *configuration_items
    ) -> DeleteConfigurationItemsReply:
        """Delete given configuration items from VPP and from VPP extended client.

        In case of problems deleting it from VPP, configuration item is not deleted from VPP extended client and
        its state changes to BLOCKED_BY_REMOVE_APPLY_FAILURE.

        Full cascade delete is not supported. That means that when configuration item that is not provided as input
        for this delete method depends on configuration item that should be deleted (=is input for this delete method)
        then it will not be deleted. However, if dependency is detected between input configuration items, their delete
        order is accordingly reordered. This means that if you provide all dependent configuration items as input, then
        everything will be removed in correct order and no leftover configuration items will be hanging (i.e. deleting
        only wireguard interface will not delete also dependent wireguard peers, but if you provide also peers in input
        then reordering will remove peers first and then wireguard interface and everything is correctly deleted)
        """
        # TODO full cascade removal of configuration is unsupported! (it should not be needed for failover case)
        if not configuration_items:
            # nothing to delete, but that nothing is deleted successfully
            return DeleteConfigurationItemsReply()

        # check for unknown config items
        record_configs = [
            record.configuration_item for record in self._configuration_item_records
        ]
        unknown_configs = [
            configuration_item
            for configuration_item in configuration_items
            if configuration_item not in record_configs
        ]
        if unknown_configs:
            return DeleteConfigurationItemsReply(
                # TODO move unknown items to processing error, make it Error instead of string?
                processing_error="Can't delete configuration item that was not previously added",
                vpp_apply_success=False,
                vpp_apply_attempted_items=unknown_configs,
            )

        # reordering of input configuration items according to dependencies (inverse order to adding config items)
        def dependency_based_reorder_for_delete(records):
            # computing correct ordering if we want to add configuration items
            ordered = []
            unordered = records.copy()
            while unordered:
                labels_from_unordered = {
                    label
                    for record in unordered
                    for label in record.handler.get_labels(record.configuration_item)
                }
                no_dependencies_records = [
                    record
                    for record in unordered
                    if not record.handler.get_dependency_labels(
                        record.configuration_item
                    ).intersection(labels_from_unordered)
                ]
                for record in no_dependencies_records:
                    ordered.append(record)
                    unordered.remove(record)

            # making from adding ordering the removal ordering
            ordered.reverse()

            return ordered

        # delete configuration items
        configuration_item_replies = []
        logger.info(
            "{}\nTransaction {} started\n{}".format(
                "=" * 30, self._transaction_counter, "=" * 30
            )
        )
        records_to_delete = [
            record
            for configuration_item in configuration_items
            for record in self._configuration_item_records
            if configuration_item == record.configuration_item
        ]
        for to_delete in dependency_based_reorder_for_delete(records_to_delete):
            try:
                reply = to_delete.handler.remove_from_vpp(to_delete.configuration_item)
            except Exception as e:
                logger.exception(
                    f"Exception raised while removing {to_delete.configuration_item}"
                )
                reply = ApplyReply(success=False, error=repr(e))
            if reply.success:
                logger.info(
                    "\tSUCCESS: removed from VPP config item {}".format(
                        repr(to_delete.configuration_item)
                    )
                )
                to_delete.metadata.update(reply.vpp_data)
                configuration_item_replies.append(
                    to_delete.get_configuration_item_reply()
                )
                self._configuration_item_records.discard(to_delete)
            else:
                logger.error(
                    "\tFAILED: removing from VPP config item {}\n\t\tREASON:{}".format(
                        repr(to_delete.configuration_item), reply.error
                    )
                )
                to_delete.state = State.BLOCKED_BY_REMOVE_APPLY_FAILURE
                configuration_item_replies.append(
                    to_delete.get_configuration_item_reply(error=reply.error)
                )

        logger.info(
            "{}\nTransaction {} ended\n{}\n".format(
                "=" * 30, self._transaction_counter, "=" * 30
            )
        )
        return DeleteConfigurationItemsReply(
            vpp_apply_success=not (
                True in [bool(reply.error) for reply in configuration_item_replies]
            ),
            vpp_apply_attempted_items=configuration_item_replies,
        )

    def update_configuration(
        self, old_configuration_item, new_configuration_item, context: str = ""
    ):
        """Server implementation for updating configuration items"""

        if self._get_handler(old_configuration_item) != self._get_handler(
            new_configuration_item
        ):
            return UpdateConfigurationItemsReply(
                processing_error="old and new configuration item must be of the same "
                "type (handled by the same handler)"
            )
        # handle generic case (=simple removal of old config item and adding new config item)
        handler = self._get_handler(old_configuration_item)
        if not isinstance(
            handler, VPPConfigUpdater
        ):  # no special update treatment -> plain delete and add
            delete_reply = self.delete_configuration(old_configuration_item)
            if (
                delete_reply.processing_error != ""
                or not delete_reply.vpp_apply_success
            ):
                return UpdateConfigurationItemsReply(
                    processing_error=delete_reply.processing_error,
                    vpp_apply_success=delete_reply.vpp_apply_success,
                    vpp_apply_attempted_items=delete_reply.vpp_apply_attempted_items,
                )
            add_reply = self.add_configuration(new_configuration_item, context=context)
            return UpdateConfigurationItemsReply(
                processing_error=add_reply.processing_error,
                all_updated_items_applied_to_vpp=add_reply.all_added_items_applied_to_vpp,
                updated_items=add_reply.added_items,
                vpp_apply_success=add_reply.vpp_apply_success,
                vpp_apply_attempted_items=add_reply.vpp_apply_attempted_items,
            )

        # handle special updates
        if isinstance(old_configuration_item, InterfaceConfigurationItem):
            # Note: the whole update was done to just update DPDK interface that is created by automatically by server
            # start -> making simplified update for this case:
            # 1. no unblocking of additional ConfigurationItems because DPDK interface exists before this update
            # 2. no blocking of this ConfigurationItem as DPDK interface doesn't depend on anything
            # -> can run handler's update_in_vpp method directly

            # check duplicate config items
            duplicate_config_items = [
                record.get_configuration_item_reply()
                for record in self._configuration_item_records
                if new_configuration_item == record.configuration_item
            ]
            if (
                duplicate_config_items
            ):  # duplicate config items are expected to fail config item addition
                return UpdateConfigurationItemsReply(
                    processing_error="Found duplicate config items.",
                    all_added_items_applied_to_vpp=False,
                    added_items=duplicate_config_items,
                )

            # update configuration
            records = [
                record
                for record in self._configuration_item_records
                if record.configuration_item == old_configuration_item
            ]
            if not records:
                return UpdateConfigurationItemsReply(
                    processing_error="Old configuration not found in server state records"
                )
            record = records[0]

            configuration_item_replies = []
            self._transaction_counter += 1
            logger.info(
                "{}\nTransaction {} started\n{}".format(
                    "=" * 30, self._transaction_counter, "=" * 30
                )
            )
            try:
                reply = handler.update_in_vpp(
                    old_configuration_item, new_configuration_item
                )
            except Exception as e:
                logger.exception(
                    f"Exception raised while updating {old_configuration_item} to "
                    f"{new_configuration_item}"
                )
                reply = ApplyReply(success=False, error=repr(e))
            if reply.success:
                logger.info(
                    "\tSUCCESS: updated VPP config item from {} to {}".format(
                        repr(old_configuration_item), repr(new_configuration_item)
                    )
                )
                record.state = State.APPLIED
                record.configuration_item = new_configuration_item
                configuration_item_replies.append(record.get_configuration_item_reply())
                record.metadata[VPP_DUMP_KEY] = handler.dump_from_vpp(
                    new_configuration_item
                )
                # Note: don't need to update sw_if_index in metadata as that didn't change
            else:
                logger.error(
                    "\tFAILED: updating VPP config item from {} to {}\n\t\tREASON:{}".format(
                        repr(old_configuration_item),
                        repr(new_configuration_item),
                        reply.error,
                    )
                )
                record.state = State.BLOCKED_BY_UPDATE_APPLY_FAILURE
                record.configuration_item = new_configuration_item
                configuration_item_replies.append(
                    record.get_configuration_item_reply(error=reply.error)
                )
            logger.info(
                "{}\nTransaction {} ended\n{}\n".format(
                    "=" * 30, self._transaction_counter, "=" * 30
                )
            )

            return UpdateConfigurationItemsReply(
                all_updated_items_applied_to_vpp=record.state == State.APPLIED,
                updated_items=[record.get_configuration_item_detail()],
                vpp_apply_success=not (
                    True in [bool(reply.error) for reply in configuration_item_replies]
                ),
                vpp_apply_attempted_items=configuration_item_replies,
            )

        # TODO implement generic code for special-written updates in handlers? (something like add_configuration
        #  but with handler.update_in_vpp(...) call, the InterfaceConfigurationItem is simplified case
        #  of the generic code)
        return UpdateConfigurationItemsReply(
            processing_error="Unhandled/Not implemented use case."
        )

    def get_configuration(self, sync_with_vpp=False) -> GetConfigurationItemsReply:
        # do synchronization with VPP
        if sync_with_vpp:
            self.sync_interfaces_with_vpp()

        # return inner state of server (with possible updated VPP state from previous step)
        return GetConfigurationItemsReply(
            items=[
                record.get_configuration_item_detail()
                for record in self._configuration_item_records
            ]
        )

    def sync_interfaces_with_vpp(self):
        for record in self._configuration_item_records:
            if isinstance(record.configuration_item, InterfaceConfigurationItem):
                record.metadata[VPP_DUMP_KEY] = record.handler.dump_from_vpp(
                    record.configuration_item
                )


class ConfigurationItemRecord:
    """Server's inner state for one configuration item"""

    def __init__(
        self, configuration_item, handler, init_state=State.RECEIVED, metadata=dict()
    ):
        self.configuration_item = configuration_item
        self.handler = handler
        self.state = init_state
        self.metadata = metadata

    def get_configuration_item_reply(self, error="") -> ConfigurationItemReply:
        """Retrieves a ConfigurationItemReply view into/copy for given record"""
        return ConfigurationItemReply(
            config=self.configuration_item,
            state=self.state,
            metadata=self.metadata,
            error=error,
        )

    def get_configuration_item_detail(self) -> ConfigurationItemDetail:
        """Retrieves a ConfigurationItemDetail view into/copy for given record"""
        return ConfigurationItemDetail(
            config=self.configuration_item, state=self.state, metadata=self.metadata
        )


def papi_wrapper(vpp_client):
    """Creates Pyro5 wrapper class that has all api methods from old PAPI client exposed for clients of PAPI++ server"""

    class Wrapper(pyro.NamedWrapper):
        @staticmethod
        def name():
            return "papi"

    for attr_name in dir(vpp_client.inner_vpp_client.api):
        if (
            attr_name.find("__") == -1
        ):  # use only public VPP API attributes (callable API functions)
            setattr(
                Wrapper,
                attr_name,
                staticmethod(getattr(vpp_client.inner_vpp_client.api, attr_name)._func),
            )
    pyro.api_expose(Wrapper)
    return Wrapper


def extended_papi_wrapper(vpp_client):
    """Creates Pyro5 wrapper class that has all the new api methods exposed for clients of PAPI++ server"""

    @pyro.api_expose
    class Wrapper(pyro.NamedWrapper):
        def __init__(self):
            # Note: synchronization could be implemented also in vpp_client because it is only one instance, it
            #  can be used different lock for readonly access,etc. ... using one lock here for simplicity and
            #  not complicating vpp_client
            self.lock = threading.Lock()

        @staticmethod
        def name():
            return "extended.papi"

        def add_configuration(
            self, *configuration_items, context: str = ""
        ) -> AddConfigurationItemsReply:
            try:
                self.lock.acquire(blocking=True)
                return vpp_client.add_configuration(
                    *configuration_items, context=context
                )
            except Exception:
                logger.exception("Exception in add_configuration")
            finally:
                self.lock.release()

        def delete_configuration(
            self, *configuration_items
        ) -> DeleteConfigurationItemsReply:
            try:
                self.lock.acquire(blocking=True)
                return vpp_client.delete_configuration(*configuration_items)
            except Exception:
                logger.exception("Exception in delete_configuration")
            finally:
                self.lock.release()

        def update_configuration(
            self, old_configuration_item, new_configuration_item, context: str = ""
        ):
            try:
                self.lock.acquire(blocking=True)
                return vpp_client.update_configuration(
                    old_configuration_item, new_configuration_item, context
                )
            except Exception:
                logger.exception("Exception in update_configuration")
            finally:
                self.lock.release()

        def get_configuration(self, sync_with_vpp=False) -> GetConfigurationItemsReply:
            try:
                self.lock.acquire(blocking=True)
                return vpp_client.get_configuration(sync_with_vpp=sync_with_vpp)
            except Exception:
                logger.exception("Exception in get_configuration")
            finally:
                self.lock.release()

    return Wrapper


def parse_args():
    import argparse

    parser = argparse.ArgumentParser(description="PAPI++ server")
    parser.add_argument(
        "--frr_agent_host",
        help="Host (IP address) of FRR agent. (Basically FRR docker container ip address). "
        "Default setting(empty string) disables connections to FRR",
        type=str,
        default="",
    )
    parser.add_argument(
        "--frr_agent_port",
        help="Port that FRR agent listen to. (Exposed port from FRR docker container for agent)",
        type=int,
        default=7777,
    )
    parser.add_argument(
        "--dhcp_socket",
        help="Path to DHCP server control unix socket. (Exposed unix socket from DHCP docker container)",
        type=str,
        default="/run/kea/control_socket_4",
    )
    # add version args
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )
    return parser.parse_args()


def main():
    args = parse_args()
    vpp_client = ExtendedVPPClient(
        "/usr/share/vpp/api/",
        "/run/vpp/api.sock",
        frr_agent_host=args.frr_agent_host,
        frr_agent_port=args.frr_agent_port,
        dhcp_socket=args.dhcp_socket,
    )  # PAPI++ instance
    pyro.run_server_loop(papi_wrapper(vpp_client), extended_papi_wrapper(vpp_client))


def cli():
    args = parse_args()
    vpp_client = ExtendedVPPClient(
        "/usr/share/vpp/api/",
        "/run/vpp/api.sock",
        frr_agent_host=args.frr_agent_host,
        frr_agent_port=args.frr_agent_port,
    )  # PAPI++ instance
    pyro.run_server_loop(papi_wrapper(vpp_client), extended_papi_wrapper(vpp_client))


if __name__ == "__main__":
    main()
